﻿using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using HealthMonitor.Core.Common;
using HealthMonitor.Util.Common.Operator;

namespace HealthMonitor.Core.Operator.Redis
{
	public class GpsCardOperatorRedisManager : IGpsCardOperatorManager
	{
		private const string GpsCardRedisKey_Insert = ".GPS_CARD_INSERT";
		private const string GpsCardRedisKey_Update = ".GPS_CARD_UPDATE";
		private const string GpsCardRedisKey_Delete = ".GPS_CARD_DELETE";

		private const string Failure_GpsCardRedisKey_Insert = ".FAILURE_GPS_CARD_INSERT";
		private const string Failure_GpsCardRedisKey_Update = ".FAILURE_GPS_CARD_UPDATE";
		private const string Failure_GpsCardRedisKey_Delete = ".FAILURE_GPS_CARD_DELETE";

		private readonly static object _syncLocker = new object();
		private readonly ILogger<GpsCardOperatorRedisManager> _logger;

		//记录当前所有队列容量（不需要精确，大概数量就可以）
		private long _existedInsert = 0;
		private long _existedUpdate = 0;
		private long _existedDelete = 0;

		public event Delegates.MonitorModeSetupHandler StartRealtimeMode = default!;
        public event Delegates.MonitorModeSetupHandler StartIdleMode = default!;

        public Dictionary<string, OperateCacheItem> PendingInsertOperators { get; private set; } = default!;
		public Dictionary<string, OperateCacheItem> PendingUpdateOperators { get; private set; } = default!;
        public Dictionary<string, OperateCacheItem> PendingDeleteOperators { get; private set; } = default!;

        public GpsCardOperatorRedisManager(ILogger<GpsCardOperatorRedisManager> logger)
		{
			_logger = logger;
		}

		public void Append(string key, OperateCacheItem item)
		{
			if (item == null) throw new ArgumentNullException($"{nameof(item)}不能为null");

			lock (_syncLocker)
			{
				//不存在则插入，否则直接覆盖
				switch (item.Type)
				{
					case DelayOperateTypeEnum.Insert:
						{
							var results = RedisHelper.StartPipe(p => p.HSet(GpsCardRedisKey_Insert, key, item).HLen(GpsCardRedisKey_Insert));
							if (results.Length != 2) throw new InvalidOperationException("Redis pipe occur errors");
							_existedInsert = (long)results[1];
						}
						break;
					case DelayOperateTypeEnum.Update:
						{
							var results = RedisHelper.StartPipe(p => p.HSet(GpsCardRedisKey_Update, key, item).HLen(GpsCardRedisKey_Update));
							if (results.Length != 2) throw new InvalidOperationException("Redis pipe occur errors");
							_existedUpdate = (long)results[1];
						}
						break;
					case DelayOperateTypeEnum.Delete:
						{
							var results = RedisHelper.StartPipe(p => p.HSet(GpsCardRedisKey_Delete, key, item).HLen(GpsCardRedisKey_Delete));
							if (results.Length != 2) throw new InvalidOperationException("Redis pipe occur errors");
							_existedDelete = (long)results[1];
						}
						break;
				}

				int limit = CoreConsts.MAX_OPERATE_COUNT / 2;
				if (_existedInsert >= limit || _existedUpdate >= limit || _existedDelete >= limit) StartRealtimeMode?.Invoke();
				else StartIdleMode?.Invoke();
			}
		}

		private Dictionary<string, OperateCacheItem>? Remove(DelayOperateTypeEnum type, int count = CoreConsts.MAX_OPERATE_COUNT)
		{
			count = Math.Min(count, CoreConsts.MAX_OPERATE_COUNT);

			try
			{
				//不能使用lock()封装，双重锁有可能导致死锁
				using (new DistributedLocker(this))
				{
					switch (type)
					{
						case DelayOperateTypeEnum.Insert:
							{
								var cached = RedisHelper.HScan<OperateCacheItem>(GpsCardRedisKey_Insert, 0, count: count);
								var dict = cached.Items.ToDictionary(e => e.field, e => e.value);
								var fields = dict.Keys.ToArray();
								if (fields.Length > 0)
								{
									var results = RedisHelper.StartPipe(p => p.HDel(GpsCardRedisKey_Insert, fields).HLen(GpsCardRedisKey_Insert));
									if (results.Length != 2) throw new InvalidOperationException("Redis pipe occur errors");
									_existedInsert = (long)results[1];
								}
								return dict;
							}
						case DelayOperateTypeEnum.Update:
							{
								var cached = RedisHelper.HScan<OperateCacheItem>(GpsCardRedisKey_Update, 0, count: count);
								var dict = cached.Items.ToDictionary(e => e.field, e => e.value);
								var fields = dict.Keys.ToArray();
								if (fields.Length > 0)
								{
									var results = RedisHelper.StartPipe(p => p.HDel(GpsCardRedisKey_Update, fields).HLen(GpsCardRedisKey_Update));
									if (results.Length != 2) throw new InvalidOperationException("Redis pipe occur errors");
									_existedUpdate = (long)results[1];
								}
								return dict;
							}
						case DelayOperateTypeEnum.Delete:
							{
								var cached = RedisHelper.HScan<OperateCacheItem>(GpsCardRedisKey_Delete, 0, count: count);
								var dict = cached.Items.ToDictionary(e => e.field, e => e.value);
								var fields = dict.Keys.ToArray();
								if (fields.Length > 0)
								{
									var results = RedisHelper.StartPipe(p => p.HDel(GpsCardRedisKey_Delete, fields).HLen(GpsCardRedisKey_Delete));
									if (results.Length != 2) throw new InvalidOperationException("Redis pipe occur errors");
									_existedDelete = (long)results[1];
								}
								return dict;
							}
					}

					int limit = CoreConsts.MAX_OPERATE_COUNT / 2;
					if (_existedInsert >= limit || _existedUpdate >= limit || _existedDelete >= limit) StartRealtimeMode?.Invoke();
					else StartIdleMode?.Invoke();
				}
			}
			catch (TimeoutException ex)
			{
				_logger.LogError($"GpsCardOperator获取并移除[{type}]操作失败，msg:{ex.Message}");
			}

			return null;
		}

		/// <summary>
		/// 通过查询键，获取待插入的数据
		/// </summary>
		/// <param name="key"></param>
		/// <returns></returns>
		public OperateCacheItem GetDelayOperator(DelayOperateTypeEnum type, string key)
		{
            switch (type)
			{
				case DelayOperateTypeEnum.Insert:
					return RedisHelper.HGet<OperateCacheItem>(GpsCardRedisKey_Insert, key);

				case DelayOperateTypeEnum.Update:
					return RedisHelper.HGet<OperateCacheItem>(GpsCardRedisKey_Update, key);

				case DelayOperateTypeEnum.Delete:
					return RedisHelper.HGet<OperateCacheItem>(GpsCardRedisKey_Delete, key);
			}
			return null!;
		}

		/// <summary>
		/// 通过查询键，更新指定缓存队列的数据
		/// </summary>
		/// <param name="type"></param>
		/// <param name="key"></param>
		/// <param name="item">新值</param>
		/// <returns>是否更新成功</returns>
		public bool ModifyDelayOperator(DelayOperateTypeEnum type, string key, OperateCacheItem item)
		{
			switch (type)
			{
				case DelayOperateTypeEnum.Insert:
					lock (_syncLocker)
					{
						if (RedisHelper.HSet(GpsCardRedisKey_Insert, key, item))
						{
							//因并发导致错误插入的，需要删除
							RedisHelper.HDel(GpsCardRedisKey_Insert, key);
							return false;
						}
						return true;
					}

				case DelayOperateTypeEnum.Update:
					lock (_syncLocker)
					{
						if (RedisHelper.HSet(GpsCardRedisKey_Update, key, item))
						{
							//因并发导致错误插入的，需要删除
							RedisHelper.HDel(GpsCardRedisKey_Update, key);
							return false;
						}
						return true;
					}

				case DelayOperateTypeEnum.Delete:
					lock (_syncLocker)
					{
						if (RedisHelper.HSet(GpsCardRedisKey_Delete, key, item))
						{
							//因并发导致错误插入的，需要删除
							RedisHelper.HDel(GpsCardRedisKey_Delete, key);
							return false;
						}
						return true;
					}
			}
			return false;
		}

		/// <summary>
		/// 通过查询键，删除指定缓存队列的数据
		/// </summary>
		/// <param name="type"></param>
		/// <param name="key"></param>
		public void RemoveDelayOperator(DelayOperateTypeEnum type, string key)
		{
			switch (type)
			{
				case DelayOperateTypeEnum.Insert:
					RedisHelper.HDel(GpsCardRedisKey_Insert, key);
					break;

				case DelayOperateTypeEnum.Update:
					RedisHelper.HDel(GpsCardRedisKey_Update, key);
					break;

				case DelayOperateTypeEnum.Delete:
					RedisHelper.HDel(GpsCardRedisKey_Delete, key);
					break;
			}
		}

		/// <summary>
		/// 获取待插入的数据集
		/// </summary>
		/// <returns></returns>
		public Dictionary<string, OperateCacheItem> GetBulkInsertOperators()
		{
			var operators = Remove(DelayOperateTypeEnum.Insert);
			PendingInsertOperators = operators!;
			return operators!;
		}

		/// <summary>
		/// 获取待更新的数据集
		/// </summary>
		/// <returns></returns>
		public Dictionary<string, OperateCacheItem> GetBulkUpdateOperators()
		{
			var operators = Remove(DelayOperateTypeEnum.Update);
			PendingUpdateOperators = operators!;
			return operators!;
		}

		/// <summary>
		/// 获取待删除的数据集
		/// </summary>
		/// <returns></returns>
		public Dictionary<string, OperateCacheItem> GetBulkDeleteOperators()
		{
			var operators = Remove(DelayOperateTypeEnum.Delete);
			PendingDeleteOperators = operators!;
			return operators!;
		}

		public void PushFailureInsertOperators(IEnumerable<OperateCacheItem> operators)
		{
			if (operators == null || operators.Count() == 0) return;
			RedisHelper.RPush<OperateCacheItem>(Failure_GpsCardRedisKey_Insert, operators.ToArray());
		}

		public void PushFailureUpdateOperators(IEnumerable<OperateCacheItem> operators)
		{
			if (operators == null || operators.Count() == 0) return;
			RedisHelper.RPush<OperateCacheItem>(Failure_GpsCardRedisKey_Update, operators.ToArray());
		}

		public void PushFailureDeleteOperators(IEnumerable<OperateCacheItem> operators)
		{
			if (operators == null || operators.Count() == 0) return;
			RedisHelper.RPush<OperateCacheItem>(Failure_GpsCardRedisKey_Delete, operators.ToArray());
		}

		public IEnumerable<OperateCacheItem> PopFailureInsertOperators(int count)
		{
			if (count <= 0) throw new ArgumentOutOfRangeException(nameof(count));

			var results = RedisHelper.StartPipe(p => p.LRange(Failure_GpsCardRedisKey_Insert, 0, count - 1).LTrim(Failure_GpsCardRedisKey_Insert, count, -1));
			if (results.Length != 2) throw new InvalidOperationException("Redis pipe occur errors");
			var operators = (results[0] as string[])!.Select(e => JsonConvert.DeserializeObject<OperateCacheItem>(e));
			return operators;
		}

		public IEnumerable<OperateCacheItem> PopFailureUpdateOperators(int count)
		{
			if (count <= 0) throw new ArgumentOutOfRangeException(nameof(count));

			var results = RedisHelper.StartPipe(p => p.LRange(Failure_GpsCardRedisKey_Update, 0, count - 1).LTrim(Failure_GpsCardRedisKey_Update, count, -1));
			if (results.Length != 2) throw new InvalidOperationException("Redis pipe occur errors");
			var operators = (results[0] as string[])!.Select(e => JsonConvert.DeserializeObject<OperateCacheItem>(e));
			return operators;
		}

		public IEnumerable<OperateCacheItem> PopFailureDeleteOperators(int count)
		{
			if (count <= 0) throw new ArgumentOutOfRangeException(nameof(count));

			var results = RedisHelper.StartPipe(p => p.LRange(Failure_GpsCardRedisKey_Delete, 0, count - 1).LTrim(Failure_GpsCardRedisKey_Delete, count, -1));
			if (results.Length != 2) throw new InvalidOperationException("Redis pipe occur errors");
			var operators = (results[0] as string[])!.Select(e => JsonConvert.DeserializeObject<OperateCacheItem>(e));
			return operators;
		}
	}
}
